/*
 * RSP Compiler plug in for Project 64 (A Nintendo 64 emulator).
 *
 * (c) Copyright 2001 jabo (jabo@emulation64.com) and
 * zilmar (zilmar@emulation64.com)
 *
 * pj64 homepage: www.pj64.net
 * 
 * Permission to use, copy, modify and distribute Project64 in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Project64 is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Project64 or software derived from Project64.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so if they want them.
 *
 */

#ifdef WIN32
# include <windows.h>
#else
# include "winlnxdefs.h"
# include <string.h>
#endif
#include <stdio.h>
#include <float.h>
#include "Rsp.h"
#include "Cpu.h"
#include "RSP Registers.h"
#include "RSP Command.h"
#include "memory.h"
#include "OpCode.h"
#include "log.h"
#include "Profiling.h"
#include "breakpoint.h"
#include "X86.h"
#include <SDL/SDL.h>

UDWORD EleSpec[32], Indx[32];
OPCODE RSPOpC;
DWORD *PrgCount, NextInstruction, RSP_Running;

void * RSP_Opcode[64];
void * RSP_RegImm[32];
void * RSP_Special[64];
void * RSP_Cop0[32];
void * RSP_Cop2[32];
void * RSP_Vector[64];
void * RSP_Lc2[32];
void * RSP_Sc2[32];

void BuildInterpreterCPU(void);
void BuildRecompilerCPU(void);

#ifdef WIN32
extern HANDLE hMutex;
#endif

void SetCPU(DWORD core) {
#ifdef WIN32
	WaitForSingleObjectEx(hMutex, INFINITE, FALSE);
#endif
	CPUCore = core;
	switch (core) {
	case RecompilerCPU:
		BuildRecompilerCPU();
		break;
	case InterpreterCPU:
		BuildInterpreterCPU();
		break;
	}
#ifdef WIN32
	ReleaseMutex(hMutex);
#endif
}

void Build_RSP ( void ) {
	int i;

	SetCPU(CPUCore);
#ifndef GCC
	ResetTimerList();
#endif

	EleSpec[ 0].DW = 0;
	EleSpec[ 1].DW = 0;
	EleSpec[ 2].DW = 0;
	EleSpec[ 3].DW = 0;
	EleSpec[ 4].DW = 0;
	EleSpec[ 5].DW = 0;
	EleSpec[ 6].DW = 0;
	EleSpec[ 7].DW = 0;
	EleSpec[ 8].DW = 0;
	EleSpec[ 9].DW = 0;
	EleSpec[10].DW = 0;
	EleSpec[11].DW = 0;
	EleSpec[12].DW = 0;
	EleSpec[13].DW = 0;
	EleSpec[14].DW = 0;
	EleSpec[15].DW = 0;
	EleSpec[16].DW = 0x0001020304050607LL; /* None */
	EleSpec[17].DW = 0x0001020304050607LL; /* None */
	EleSpec[18].DW = 0x0000020204040606LL; /* 0q */
	EleSpec[19].DW = 0x0101030305050707LL; /* 1q */
	EleSpec[20].DW = 0x0000000004040404LL; /* 0h */
	EleSpec[21].DW = 0x0101010105050505LL; /* 1h */
	EleSpec[22].DW = 0x0202020206060606LL; /* 2h */
	EleSpec[23].DW = 0x0303030307070707LL; /* 3h */
	EleSpec[24].DW = 0x0000000000000000LL; /* 0 */
	EleSpec[25].DW = 0x0101010101010101LL; /* 1 */
	EleSpec[26].DW = 0x0202020202020202LL; /* 2 */
	EleSpec[27].DW = 0x0303030303030303LL; /* 3 */
	EleSpec[28].DW = 0x0404040404040404LL; /* 4 */
	EleSpec[29].DW = 0x0505050505050505LL; /* 5 */
	EleSpec[30].DW = 0x0606060606060606LL; /* 6 */
	EleSpec[31].DW = 0x0707070707070707LL; /* 7 */

	Indx[ 0].DW = 0;
	Indx[ 1].DW = 0;
	Indx[ 2].DW = 0;
	Indx[ 3].DW = 0;
	Indx[ 4].DW = 0;
	Indx[ 5].DW = 0;
	Indx[ 6].DW = 0;
	Indx[ 7].DW = 0;
	Indx[ 8].DW = 0;
	Indx[ 9].DW = 0;
	Indx[10].DW = 0;
	Indx[11].DW = 0;
	Indx[12].DW = 0;
	Indx[13].DW = 0;
	Indx[14].DW = 0;
	Indx[15].DW = 0;

	Indx[16].DW = 0x0001020304050607LL; /* None */
	Indx[17].DW = 0x0001020304050607LL; /* None */
	Indx[18].DW = 0x0103050700020406LL; /* 0q */
	Indx[19].DW = 0x0002040601030507LL; /* 1q */
	Indx[20].DW = 0x0102030506070004LL; /* 0h */
	Indx[21].DW = 0x0002030406070105LL; /* 1h */
	Indx[22].DW = 0x0001030405070206LL; /* 2h */
	Indx[23].DW = 0x0001020405060307LL; /* 3h */
	Indx[24].DW = 0x0102030405060700LL; /* 0 */
	Indx[25].DW = 0x0002030405060701LL; /* 1 */
	Indx[26].DW = 0x0001030405060702LL; /* 2 */
	Indx[27].DW = 0x0001020405060703LL; /* 3 */
	Indx[28].DW = 0x0001020305060704LL; /* 4 */
	Indx[29].DW = 0x0001020304060705LL; /* 5 */
	Indx[30].DW = 0x0001020304050706LL; /* 6 */
	Indx[31].DW = 0x0001020304050607LL; /* 7 */

	for (i = 16; i < 32; i ++) {
		int count;

		for (count = 0; count < 8; count ++) {
			Indx[i].B[count] = 7 - Indx[i].B[count];
			EleSpec[i].B[count] = 7 - EleSpec[i].B[count];
		}
		for (count = 0; count < 4; count ++) {
			BYTE Temp;
			
			Temp = Indx[i].B[count];
			Indx[i].B[count] = Indx[i].B[7 - count]; 
			Indx[i].B[7 - count] = Temp;
		}
	}

	PrgCount = RSPInfo.SP_PC_REG;
}

/******************************************************************
  Function: DoRspCycles
  Purpose:  This function is to allow the RSP to run in parrel with
            the r4300 switching control back to the r4300 once the
			function ends.
  input:    The number of cylces that is meant to be executed
  output:   The number of cycles that was executed. This value can
            be greater than the number of cycles that the RSP 
			should have performed.
			(this value is ignored if the RSP is stoped)
*******************************************************************/ 

DWORD RunInterpreterCPU(DWORD Cycles);
DWORD RunRecompilerCPU ( DWORD Cycles );


// #define THREADED
// #define ASYNC

#ifdef THREADED
SDL_sem * rspCommandSema;
SDL_sem * rspCommandCompleteSema;
SDL_Thread * rspThread;
DWORD _DoRspCycles ( DWORD Cycles );

int rspThreadFunc(void * dummy)
{
  while (1) {
    SDL_SemWait(rspCommandSema);

    _DoRspCycles(100000);

#ifndef ASYNC
    SDL_SemPost(rspCommandCompleteSema);
#endif
  }
  return 0;
}

void rspPostCommand()
{
  SDL_SemPost(rspCommandSema);
#ifndef ASYNC
  SDL_SemWait(rspCommandCompleteSema);
#endif
}

void rspCreateThread()
{
  if (!rspCommandSema) {
    rspCommandSema = SDL_CreateSemaphore(0);
    rspCommandCompleteSema = SDL_CreateSemaphore(0);
  }
  if (!rspThread) {
    printf("Creating rsp thread\n");
    rspThread = SDL_CreateThread(rspThreadFunc, 0);
  }
}

__declspec(dllexport) DWORD DoRspCycles ( DWORD Cycles )
{
  rspCreateThread();
  rspPostCommand();
  return Cycles;
}
#endif


#ifdef THREADED
DWORD _DoRspCycles ( DWORD Cycles ) {
#else
__declspec(dllexport) DWORD DoRspCycles ( DWORD Cycles ) {
#endif
	extern BOOL AudioHle, GraphicsHle;
	DWORD TaskType = *(DWORD*)(RSPInfo.DMEM + 0xFC0);

#ifdef THREADED
  AudioHle = TRUE;
#endif
  
	if (TaskType == 1 && GraphicsHle) {
		if (RSPInfo.ProcessDList != NULL) {
			RSPInfo.ProcessDList();
		}
		*RSPInfo.SP_STATUS_REG |= (0x0203 );
		if ((*RSPInfo.SP_STATUS_REG & SP_STATUS_INTR_BREAK) != 0 ) {
			*RSPInfo.MI_INTR_REG |= R4300i_SP_Intr;
			RSPInfo.CheckInterrupts();
		}

		*RSPInfo.DPC_STATUS_REG &= ~0x0002;
		return Cycles;
	} else if (TaskType == 2 && AudioHle) {
#ifndef THREADED
		if (RSPInfo.ProcessAList != NULL) {
			RSPInfo.ProcessAList();
		}
#endif
		*RSPInfo.SP_STATUS_REG |= (0x0203 );
		if ((*RSPInfo.SP_STATUS_REG & SP_STATUS_INTR_BREAK) != 0 ) {
			*RSPInfo.MI_INTR_REG |= R4300i_SP_Intr;
			RSPInfo.CheckInterrupts();
		}
		return Cycles;
	} else if (TaskType == 7) {
    //printf("showcfb\n");
		RSPInfo.ShowCFB();		
	}

	/* 
	*RSPInfo.SP_STATUS_REG |= (0x0203 );
	if ((*RSPInfo.SP_STATUS_REG & SP_STATUS_INTR_BREAK) != 0 ) {
		*RSPInfo.MI_INTR_REG |= R4300i_SP_Intr;
		RSPInfo.CheckInterrupts();
	}
	return Cycles;
	*/

#if !defined(EXTERNAL_RELEASE)
	if (Profiling) {
		StopTimer();
		if (!IndvidualBlock) { StartTimer("RSP Running"); }
	}
#endif

  //extern int dump_dma;
  if (TaskType == 1) {
//     dump_dma = 1;

//     DWORD dlist_start = *(DWORD*)(RSPInfo.DMEM+0xFF0);
//     DWORD dlist_length = *(DWORD*)(RSPInfo.DMEM+0xFF4);

//     printf("dlist %x --> %x\n", dlist_start, dlist_length);

//     int i;
//     for (i=0xFC0 ; i<0x1000; i+=4)
//       printf("%x = %x\n", i, *(DWORD*)(RSPInfo.DMEM+i));
  }

#ifdef WIN32
	WaitForSingleObjectEx(hMutex, INFINITE, FALSE);
#endif
	switch (CPUCore) {
	case RecompilerCPU:
		RunRecompilerCPU(Cycles);
		break;
	case InterpreterCPU:
		RunInterpreterCPU(Cycles);
		break;
	}
#ifdef WIN32
	ReleaseMutex(hMutex);
#endif

  //if (TaskType == 1) dump_dma = 0;

#if !defined(EXTERNAL_RELEASE)
	if (Profiling) {
		StopTimer();
		StartTimer("r4300i code");
	}
#endif

	return Cycles;
}
